@@ -1,6 +1,7 @@
-from .utils import INF, get_bound, uniq, fsplit, drange
-from .utils import bbox2str, matrix2str, apply_matrix_pt
+from itertools import combinations
 
+from .utils import (INF, get_bound, uniq, fsplit, drange, bbox2str, matrix2str, apply_matrix_pt,
+    trailiter)
 
 class IndexAssigner:
 
@@ -19,7 +20,7 @@ class IndexAssigner:
 class LAParams:
 
     def __init__(self, line_overlap=0.5, char_margin=2.0, line_margin=0.5, word_margin=0.1,
-            boxes_flow=0.5, detect_vertical=False, all_texts=False):
+            boxes_flow=0.5, detect_vertical=False, all_texts=False, paragraph_indent=None):
         self.line_overlap = line_overlap
         self.char_margin = char_margin
         self.line_margin = line_margin
@@ -27,6 +28,10 @@ class LAParams:
         self.boxes_flow = boxes_flow
         self.detect_vertical = detect_vertical
         self.all_texts = all_texts
+        # If this setting is not None, horizontal text boxes will be split by paragraphs, using
+        # the indent of their first line for the split. The numerical argument is the treshold that
+        # the line's x-pos must reach to be considered "indented".
+        self.paragraph_indent = paragraph_indent
 
     def __repr__(self):
         return ('<LAParams: char_margin=%.1f, line_margin=%.1f, word_margin=%.1f all_texts=%r>' %
@@ -285,21 +290,40 @@ class LTTextLineHorizontal(LTTextLine):
 
     def __init__(self, word_margin):
         LTTextLine.__init__(self, word_margin)
-        self._x1 = +INF
+        self._last_char = None
+        # for height average
+        self._charheight_sum = 0
+        self._charcount = 0
 
     def add(self, obj):
-        if isinstance(obj, LTChar) and self.word_margin:
+        assert isinstance(obj, LTChar)
+        # if we're at the beginning of the line, we consider ourselves 'inspace'
+        inspace = (self._last_char is None) or (self._last_char.get_text() == ' ') or (obj.get_text() == ' ')
+        if self.word_margin and not inspace:
             margin = self.word_margin * obj.width
-            if self._x1 < obj.x0-margin:
+            if self._last_char.x1 < obj.x0-margin:
                 LTContainer.add(self, LTAnon(' '))
-        self._x1 = obj.x1
+        self._last_char = obj
+        self._charheight_sum += obj.height
+        self._charcount += 1
         LTTextLine.add(self, obj)
 
     def find_neighbors(self, plane, ratio):
         h = ratio*self.height
         objs = plane.find((self.x0, self.y0-h, self.x1, self.y1+h))
-        return [ obj for obj in objs if isinstance(obj, LTTextLineHorizontal) ]
+        ACCEPTABLE_DIFF = 3
+        acceptable = lambda obj: isinstance(obj, LTTextLineHorizontal) and\
+            abs(obj.avg_charheight - self.avg_charheight) < ACCEPTABLE_DIFF
+        return [obj for obj in objs if acceptable(obj)]
     
+    @property
+    def avg_charheight(self):
+        if self._charcount:
+            return self._charheight_sum / self._charcount
+        else:
+            return 0
+    
+
 class LTTextLineVertical(LTTextLine):
 
     def __init__(self, word_margin):
@@ -336,12 +360,73 @@ class LTTextBox(LTTextContainer):
 
 class LTTextBoxHorizontal(LTTextBox):
     
+    def __init__(self):
+        LTTextBox.__init__(self)
+        self._avg_lineheight = None
+    
+    def add(self, obj):
+        LTTextBox.add(self, obj)
+        self._avg_lineheight = None
+    
     def analyze(self, laparams):
         LTTextBox.analyze(self, laparams)
-        self._objs = sorted(self._objs, key=lambda obj: -obj.y1)
-
+        self._sort_lines()
+    
+    def _pos_in_box(self, obj):
+        if self._avg_lineheight is None:
+            self._avg_lineheight = sum(o.height for o in self._objs) / len(self._objs)
+        x = obj.x0 - self.x0
+        y = self.y1 - obj.y1
+        # gridy is a y pos rounded using half the average line height. This way, we can be
+        # confident that lines that have almost the same Y-pos will have the same gridy
+        gridy = round(y / (self._avg_lineheight / 2))
+        return x, y, gridy
+    
+    def _sort_lines(self):
+        # Sorting lines in our textbox is not so easy. It's possible that we get some lines that
+        # are obviously the same, but one of them is slightly higher or lower. In these cases,
+        # simply sorting by Y-pos will be wrong. That's why we take the average line height to
+        # "snap" our y-pos to some kind of grid. Then we sort by "snapped" ypos, using X pos as
+        # a tie breaker.
+        def sortkey(obj):
+            x, y, gridy = self._pos_in_box(obj)
+            return (gridy, x)
+        
+        self._objs = sorted(self._objs, key=sortkey)
+    
     def get_writing_mode(self):
         return 'lr-tb'
+    
+    def paragraphs(self, indent_treshold):
+        # Check if some lines in the box are indented and if yes, split our textbox in multiple
+        # paragraphs and return the result.
+        self._sort_lines()
+        prev_lines_xpos = []
+        paragraphs = []
+        current_paragraph = LTTextBoxHorizontal()
+        prevgridy = None
+        for obj in self._objs:
+            x, y, gridy = self._pos_in_box(obj)
+            if gridy != prevgridy:
+                if (x > indent_treshold) and current_paragraph:
+                    paragraphs.append(current_paragraph)
+                    current_paragraph = LTTextBoxHorizontal()
+                prev_lines_xpos.append(x)
+                prevgridy = gridy
+            current_paragraph.add(obj)
+        if current_paragraph:
+            paragraphs.append(current_paragraph)
+        # Now, it's possible that we have box like a title that has no paragraph at all but that the
+        # algo above would mistakenly see as one-paragraph-per line (a centered, multi-line title)
+        # we must thus check if most of our lines are "normal" lines (x < 1)
+        normal_line_count = len([x for x in prev_lines_xpos if x < 1])
+        if normal_line_count < len(prev_lines_xpos) / 2:
+            return [self]
+        if len(paragraphs) > 1:
+            return paragraphs
+        else:
+            return [self]
+    
 
 class LTTextBoxVertical(LTTextBox):
 
@@ -387,63 +472,61 @@ class LTLayoutContainer(LTContainer):
     def get_textlines(self, laparams, objs):
         obj0 = None
         line = None
-        for obj1 in objs:
-            if obj0 is not None:
-                k = 0
-                if (obj0.is_compatible(obj1) and obj0.is_voverlap(obj1) and 
-                    min(obj0.height, obj1.height) * laparams.line_overlap < obj0.voverlap(obj1) and
-                    obj0.hdistance(obj1) < max(obj0.width, obj1.width) * laparams.char_margin):
-                    # obj0 and obj1 is horizontally aligned:
-                    #
-                    #   +------+ - - -
-                    #   | obj0 | - - +------+   -
-                    #   |      |     | obj1 |   | (line_overlap)
-                    #   +------+ - - |      |   -
-                    #          - - - +------+
-                    #
-                    #          |<--->|
-                    #        (char_margin)
-                    k |= 1
-                if (laparams.detect_vertical and
-                    obj0.is_compatible(obj1) and obj0.is_hoverlap(obj1) and 
-                    min(obj0.width, obj1.width) * laparams.line_overlap < obj0.hoverlap(obj1) and
-                    obj0.vdistance(obj1) < max(obj0.height, obj1.height) * laparams.char_margin):
-                    # obj0 and obj1 is vertically aligned:
-                    #
-                    #   +------+
-                    #   | obj0 |
-                    #   |      |
-                    #   +------+ - - -
-                    #     |    |     | (char_margin)
-                    #     +------+ - -
-                    #     | obj1 |
-                    #     |      |
-                    #     +------+
-                    #
-                    #     |<-->|
-                    #   (line_overlap)
-                    k |= 2
-                if ( (k & 1 and isinstance(line, LTTextLineHorizontal)) or
-                     (k & 2 and isinstance(line, LTTextLineVertical)) ):
+        for obj0, obj1 in trailiter(objs, skipfirst=True):
+            k = 0
+            if (obj0.is_compatible(obj1) and obj0.is_voverlap(obj1) and 
+                min(obj0.height, obj1.height) * laparams.line_overlap < obj0.voverlap(obj1) and
+                obj0.hdistance(obj1) < max(obj0.width, obj1.width) * laparams.char_margin):
+                # obj0 and obj1 is horizontally aligned:
+                #
+                #   +------+ - - -
+                #   | obj0 | - - +------+   -
+                #   |      |     | obj1 |   | (line_overlap)
+                #   +------+ - - |      |   -
+                #          - - - +------+
+                #
+                #          |<--->|
+                #        (char_margin)
+                k |= 1
+            if (laparams.detect_vertical and
+                obj0.is_compatible(obj1) and obj0.is_hoverlap(obj1) and 
+                min(obj0.width, obj1.width) * laparams.line_overlap < obj0.hoverlap(obj1) and
+                obj0.vdistance(obj1) < max(obj0.height, obj1.height) * laparams.char_margin):
+                # obj0 and obj1 is vertically aligned:
+                #
+                #   +------+
+                #   | obj0 |
+                #   |      |
+                #   +------+ - - -
+                #     |    |     | (char_margin)
+                #     +------+ - -
+                #     | obj1 |
+                #     |      |
+                #     +------+
+                #
+                #     |<-->|
+                #   (line_overlap)
+                k |= 2
+            if ( (k & 1 and isinstance(line, LTTextLineHorizontal)) or
+                 (k & 2 and isinstance(line, LTTextLineVertical)) ):
+                line.add(obj1)
+            elif line is not None:
+                yield line
+                line = None
+            else:
+                if k == 2:
+                    line = LTTextLineVertical(laparams.word_margin)
+                    line.add(obj0)
                     line.add(obj1)
-                elif line is not None:
+                elif k == 1:
+                    line = LTTextLineHorizontal(laparams.word_margin)
+                    line.add(obj0)
+                    line.add(obj1)
+                else:
+                    line = LTTextLineHorizontal(laparams.word_margin)
+                    line.add(obj0)
                     yield line
                     line = None
-                else:
-                    if k == 2:
-                        line = LTTextLineVertical(laparams.word_margin)
-                        line.add(obj0)
-                        line.add(obj1)
-                    elif k == 1:
-                        line = LTTextLineHorizontal(laparams.word_margin)
-                        line.add(obj0)
-                        line.add(obj1)
-                    else:
-                        line = LTTextLineHorizontal(laparams.word_margin)
-                        line.add(obj0)
-                        yield line
-                        line = None
-            obj0 = obj1
         if line is None:
             line = LTTextLineHorizontal(laparams.word_margin)
             line.add(obj0)
@@ -472,7 +555,12 @@ class LTLayoutContainer(LTContainer):
             box = boxes[line]
             if box in done: continue
             done.add(box)
-            yield box
+            if laparams.paragraph_indent:
+                paragraphs = box.paragraphs(laparams.paragraph_indent)
+                for p in paragraphs:
+                    yield p
+            else:
+                yield box
 
     def group_textboxes(self, laparams, boxes):
         def dist(obj1, obj2):
@@ -503,13 +591,10 @@ class LTLayoutContainer(LTContainer):
             return objs.difference((obj1,obj2))
         # XXX this still takes O(n^2)  :(
         dists = []
-        for i in range(len(boxes)):
-            obj1 = boxes[i]
-            for j in range(i+1, len(boxes)):
-                obj2 = boxes[j]
-                dists.append((0, dist(obj1, obj2), obj1, obj2))
+        for obj1, obj2 in combinations(boxes, 2):
+            dists.append((0, dist(obj1, obj2), obj1, obj2))
         # we sort by dist and our tuple is (c,dist,obj1,obj2)
-        sortkey = lambda tup: tup[1]
+        sortkey = lambda tup: tup[:2]
         dists.sort(key=sortkey)
         plane = Plane(boxes)
         while dists:
@@ -517,17 +602,14 @@ class LTLayoutContainer(LTContainer):
             if c == 0 and isany(obj1, obj2):
                 dists.append((1,d,obj1,obj2))
                 continue
-            if (isinstance(obj1, LTTextBoxVertical) or
-                isinstance(obj1, LTTextGroupTBRL) or
-                isinstance(obj2, LTTextBoxVertical) or
-                isinstance(obj2, LTTextGroupTBRL)):
+            if (isinstance(obj1, (LTTextBoxVertical, LTTextGroupTBRL)) or
+                isinstance(obj2, (LTTextBoxVertical, LTTextGroupTBRL))):
                 group = LTTextGroupTBRL([obj1,obj2])
             else:
                 group = LTTextGroupLRTB([obj1,obj2])
             plane.remove(obj1)
             plane.remove(obj2)
-            dists = [ (c,d,o1,o2) for (c,d,o1,o2) in dists
-                      if o1 in plane and o2 in plane ]
+            dists = [(c,d,o1,o2) for (c,d,o1,o2) in dists if o1 in plane and o2 in plane]
             for other in plane:
                 dists.append((0, dist(group,other), group, other))
             dists.sort(key=sortkey)
